/*
 * iflib_act.c - Re-arrange expressions containing policing and other actions
 *
 * Written 2001 by Werner Almesberger
 * Copyright 2001 EPFL-ICA, Network Robots
 */

/*
 * The canonical form for actions (i.e. policing and decisions) is to be at
 * the end of && chains, after all static matches. iflib_actions bubbles all
 * non-actions up the tree.
 */

/*
 * By generously combining matches from all over the place, we may generate
 * redundant or conflicting rules here. This is left to the user (i.e. if_ext)
 * to sort out.
 */


#include <assert.h>

#include "config.h"
#include "data.h"
#include "op.h"
#include "iflib.h"


/*
 * A sub-tree is self_contained if it selects a class, or if any failed
 * conformance decisions lead to a self_contained subtree.
 */

static int self_contained(DATA d)
{
    if (!d.op) return d.type == dt_class_if;
    if (d.op->dsc == &op_conform) return 0;
    if (d.op->dsc == &op_count) return 1;
    if (d.op->dsc == &op_logical_and) return self_contained(d.op->a);
    if (d.op->dsc == &op_logical_or)
	return self_contained(d.op->a) || self_contained(d.op->b);
    return 0;
}


/*
 * Transforms  A && M  ->  M && A
 */

static void bubble_and(DATA *d)
{
    DATA tmp;

    assert(d->op && d->op->dsc == &op_logical_and);
    if (!action_tree(d->op->a) || action_tree(d->op->b)) return;
	/* already in canonical form */
    tmp = d->op->a;
    d->op->a = d->op->b;
    d->op->b = tmp;
}


/*
 * Transforms an expression of the type  (M && P) || X  where M are static
 * matches, and P are non-self-contained policing actions, by attaching a
 * copy of X at the place of P, and replacing at each transition Mi->Ai
 * from matches to an action tree (i.e. policing or decisions), Ai with
 * (P || Ai)
 *
 * E.g.  (M1 && P1) || (M2 && P2) || D  would become
 * (M1 && ((M2 && (P1 || P2)) || (P1 || D)) || (M2 && P2) || D
 *
 * Now, || and && still remain lists, but this transformation might disturb
 * the "|| over &&" structure for matches, which would throw off later
 * stages. Therefore, we merge the || chain of X (after inserting P) back on
 * top by extracting each element E and changing it to (M && E).
 */

static void attach_prev(DATA *d,DATA attach)
{
    if (action_tree(*d)) {
	*d = op_binary(&op_logical_or,data_clone(attach),*d);
	return;
    }
    if (!d->op) return;
    attach_prev(&d->op->a,attach);
    attach_prev(&d->op->b,attach);
    attach_prev(&d->op->c,attach);
}


static void add_precondition(DATA *d,const DATA *cond,DATA *cut)
{
    *cut = *d;
    *d = data_clone(*cond);
}


static void precondition_or(DATA *d,const DATA *cond,DATA *cut)
{
    while (d->op && d->op->dsc == &op_logical_or) {
	add_precondition(&d->op->a,cond,cut);
	d = &d->op->b;
    }
    add_precondition(d,cond,cut);
}


static void bubble_or(DATA *d)
{
    DATA *cut,p,x,x_cond;

    assert(d->op && d->op->dsc == &op_logical_or);
    for (cut = &d->op->a; cut->op && cut->op->dsc == &op_logical_and;
      cut = &cut->op->b)
	if (action_tree(*cut)) break;
    if (self_contained(*cut)) return;
	/* already in canonical form */
    p = *cut;
    x = data_clone(d->op->b);
    x_cond = data_clone(x);
    if (debug > 1) debug_expr("x_cond BEFORE precondition_or",x_cond);
    precondition_or(&x_cond,&d->op->a,cut);
    if (debug > 1) debug_expr("x_cond AFTER precondition_or",x_cond);
    attach_prev(&x_cond,p);
    if (debug > 1) debug_expr("x_cond AFTER attach_prev",x_cond);
    data_destroy(*d);
    *d = x_cond;
    while (d->op && d->op->dsc == &op_logical_or) d = &d->op->b;
    *d = op_binary(&op_logical_or,*d,x);
}


static void push_action(DATA *d)
{
    if (!d->op) return;
    push_action(&d->op->a);
    push_action(&d->op->b);
    push_action(&d->op->c);
    if (d->op->dsc == &op_logical_and) bubble_and(d);
    else if (d->op->dsc == &op_logical_or) bubble_or(d);
}


void iflib_actions(DATA *d)
{
    push_action(d);
    debug_expr("AFTER push_action",*d);
}
